require_relative 'manager'
require_relative 'debug_logger'
require_relative 'httpsender'
require_relative 'constants'

module Coralogix
    class CoralogixLogger
        # Constructor. 
        #
        # @param name    -   logger name.
        def initialize private_key, app_name, sub_system_name, debug=false, type_name="Centralized Ruby", force_compression=false, proxy={}, external_logger=nil
            self.debug_mode=debug            
            @category = CORALOGIX_CATEGORY
            @manager = LoggerManager.new proxy
            unless external_logger.nil?
                DebugLogger.enabled = true
                DebugLogger.external = true
                DebugLogger.logger = external_logger
            end
            configure(private_key, app_name, sub_system_name, type_name, force_compression)
        end

        # A getter for debug_mode. 
        # Default value is false.
        # When set to true the coralogix logger will print output messages to a console and a file.
        #
        # @return [boolean]    -   true or false. (Default is false)
        def debug_mode?
            DebugLogger.debug_mode
        end

        # A setter for debug_mode.  
        # Default value is false.
        # When set to true the coralogix logger will print output messages to a console and a file.
        #
        # @param value    -   true or false. (Default is false)
        def debug_mode=(value)
            DebugLogger.debug_mode=value
        end

        # A setter for disable_proxy. 
        # By default HTTP object will use proxy environment variable if exists. In some cases this migh be an issue
        # When set to false the HTTP object will ignore any proxy.
        #
        # @param value    -   true or false. (Default is false)
        def disable_proxy=(value)
            @manager.disable_proxy=value
        end

        # Configure coralogix logger with customer specific values
        #
        # @param private_key           -   private key
        # @param app_name              -   application name
        # @param sub_system            -   sub system name
        # @param type_name             -   type name
        # @param force_compression     -   when true, will not send the raw data if data compression failed 
        # @return [boolean] return a true or false.
        def configure private_key, app_name, sub_system, type_name, force_compression
            private_key = (private_key.nil? || private_key.to_s.strip.empty?) ? FAILED_PRIVATE_KEY : private_key
            app_name = (app_name.nil? || app_name.to_s.strip.empty?) ? NO_APP_NAME : app_name
            sub_system = (sub_system.nil? || sub_system.to_s.strip.empty?) ? NO_SUB_SYSTEM : sub_system
            @manager.set_logger_type_name type_name
            @manager.force_compression = force_compression
            @manager.configure(:privateKey => private_key, :applicationName => app_name, :subsystemName => sub_system) unless @manager.configured
        end


        # Flush all messages in buffer and send them immediately on the current thread.
        def flush
            @manager.flush
        end

        # Log a message. 
        #
        # @param severity    -   log severity
        # @param message     -   log message
        # @param category    -   log category
        # @param className   -   log class name
        # @param methodName  -   log method name
        # @param threadId    -   log thread id
        def log severity, message, category=@category, named_args = {}

            className = named_args.fetch(:className,"")
            methodName = named_args.fetch(:methodName,"")
            threadId = named_args.fetch(:threadId,Thread.current.object_id.to_s)

            @manager.add_logline message, severity, category, :className => className, :methodName => methodName, :threadId => threadId
        end 

        # Create log methods for each severity. 
        # This is a ruby thing. If you are writing in other languages just create a method for each severity.
        # For instance, for info severity it will create a method:
        # def info message, category: @category, className: "", methodName: "", threadId: ""
        SEVERITIES.keys.each do |severity|
            define_method("#{severity}") do |message, category=@category, named_args = {}|

                className = named_args.fetch(:className,"")
                methodName = named_args.fetch(:methodName,"")
                threadId = named_args.fetch(:threadId,Thread.current.object_id.to_s)
                timestamp = named_args.fetch(:timestamp, nil)
                @manager.add_logline message, SEVERITIES["#{__method__}".to_sym], category, :className => className, :methodName => methodName, :threadId => threadId, :timestamp => timestamp
            end  
        end
    end
end
